// BF92--machine-language Brainf*ck interpreter for the TI92/89
// Created 4/4/01; 10:47:43 PM

#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization

#define SAVE_SCREEN           // Save/Restore LCD Contents

#define DEF_TAPE 128
#define TAPE_INC 128

#include <tigcclib.h>         // Include All Header Files

short _ti92plus;              // Produce .9XZ File
short _ti89;		              // Produce .89Z File

// Main Function
void _main(void)
{

	// bf92 runs code out of a string, passed as the first argument.
	
	ESI argptr;
	char *str;
	char *pc;
	HANDLE tapeHandle;
	unsigned short tapeOffset;
	unsigned short tapeLen;
	char *tapeStart;
	char run;
	int temp;
	
	// the program starts at str and ends with a null character.
	
	
	InitArgPtr(argptr); // obtain pointer to argument stack
	FontSetSys(F_6x8);
	if(ArgCount()==1 && GetArgType(argptr)==STR_TAG) { // it's a string
		clrscr(); // prepare screen for any output
		
		pc=str=GetStrnArg(argptr); // get pointer to program string
		
		// create tape
		tapeHandle = HeapAlloc(DEF_TAPE); // allocate tape
																			
		if(tapeHandle == H_NULL) { // couldn't allocate tape
			puts("Could not allocate tape!\n(not enough free RAM)");
		} else { // tape allocated
			tapeStart = (char*)HeapDeref(tapeHandle);
			tapeOffset=0;
			tapeLen = HeapSize(tapeHandle);

			OSEnableBreak(); // allow user to abort BF loops
			memset(tapeStart, 0, tapeLen); // clear tape
			run=1;
			while(run) {
				switch(*pc) {
					case '<':
//						printf("<");
						if(tapeOffset > 0) {
							tapeOffset--;
						}
						break;
					case '>':
//						printf(">");
						if(tapeOffset < tapeLen) {
							tapeOffset++;
						} else {
//							printf("resizing heap to %u",(tapeLen+TAPE_INC));
							if(HeapRealloc(tapeHandle, tapeLen+TAPE_INC)!=H_NULL){
								tapeStart = (char*)HeapDeref(tapeHandle);
								tapeLen = HeapSize(tapeHandle);
								memset(tapeStart+tapeLen-TAPE_INC,0,TAPE_INC);
//								printf("Heap size:%u\n",tapeLen);
								tapeOffset++;
							} else {
//								printf("\nHeap reallocation failed!");
								run=0;
							}
						}
						break;
					case '+':
//						printf("+");
						(*(tapeStart+tapeOffset))++;
						break;
					case '-':
//						printf("-");
						(*(tapeStart+tapeOffset))--;
						break;
					case '.':
//						printf(".");
						putchar(*(tapeStart+tapeOffset));
						break;
					case ',':
//						printf(",");
						*(tapeStart+tapeOffset) = (char)ngetchx();
						break;
					case '[':
//						printf("[");
						if(*(tapeStart+tapeOffset) == 0) { // find corresponding ]
							temp = 1; // temp will be used for the bracket count
							while(temp != 0) {
								pc++;
								if(*pc == '[') temp++;
								if(*pc == ']') temp--;
								if(*pc == 0) run=0;
							}
						}
						break;
					case ']':
//						printf("]");
						if(*(tapeStart+tapeOffset) != 0) { // find corresponding [
							temp = -1; // temp will be used for the bracket count
							while(temp != 0) {
								pc--;
								if(*pc == '[') temp++;
								if(*pc == ']') temp--;
								if(pc < str) run=0;
							}
							pc--;
						}
						break;
				} //switch
				pc++;
				if(OSCheckBreak()==TRUE || *pc==0) run=0;
			} //while running

			// closing: free the heap
			HeapFree(tapeHandle);
		} //if (tape allocation)
		puts("\n");
		FontSetSys(F_4x6);
		puts("\n\nFinished.  Press any key for the shell.");
		ngetchx(); // wait for keystroke
	} else { // it's not a string, and thus illegal.
		HANDLE errorbox=DialogNewSimple(150,70);
		DialogAddTitle(errorbox,"Invalid Argument",BT_NONE,BT_OK);
		DialogAddText(errorbox,3,15,"Syntax:");
		DialogAddText(errorbox,3,24,"bf92(program_string)");
		DialogDo(errorbox,CENTER,CENTER,NULL,NULL);
/*		clrscr();
		DrawStr(50,50,"Syntax:",A_REPLACE);
		DrawStr(50,59,"bf92(program-string)",A_REPLACE);*/
	}
	
}
